/*
 * Copyright (C) 2010 The Android-x86 Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author: Yi Sun <beyounn@gmail.com>
 */

// ****************** Copyright (c) 2014 by Gigaset Communications GmbH ******
// Filename : EthernetService.java
// Owner    : Karthick [STC]
//-------------------------------------------------------------------------
// Definition : core of ethernet stack which is used to
//       1. manange dhcp/static ipv4 and ipv6 configuration
//       2. PC Client/PC WifiClient configurations
//       3. Ethernet states
//-------------------------------------------------------------------------
// 22-07-2013, Gigaset Communications GmbH
// Feature added to track state of ethernet connectivity
//******************************************************************************

package com.maxwell.server;

import java.net.UnknownHostException;
import com.maxwell.net.ethernet.EthernetNative;
import com.maxwell.net.ethernet.IEthernetManager;
import com.maxwell.net.ethernet.EthernetManager;
import com.maxwell.net.ethernet.EthernetStateTracker;
import com.maxwell.net.ethernet.EthernetDevInfo;
import android.provider.Settings;
import android.util.Slog;
import android.content.BroadcastReceiver;
import android.content.ContentResolver;
import android.content.Context;
import android.widget.Toast;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.SystemProperties;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.net.InetAddress;
import android.R;
import android.app.Notification;
import java.net.Inet4Address;
import java.net.Inet6Address;
import android.net.LinkAddress;
import android.net.RouteInfo;
import android.net.DhcpInfo;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.bluetooth.BluetoothHeadset;
import android.net.LinkCapabilities;
import android.net.LinkProperties;
import android.net.ConnectivityManager;
import android.net.DhcpInfoInternal;
import android.net.NetworkStateTracker;
import android.net.NetworkUtils;
import android.net.LinkAddress;
import android.net.NetworkInfo;
import android.net.NetworkInfo.DetailedState;
import android.net.LinkProperties;
import android.net.NetworkInfo.DetailedState;
import android.net.InterfaceConfiguration;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.os.Message;
import android.os.Parcel;
import android.os.ServiceManager;
import android.os.IBinder;
import android.os.INetworkManagementService;
import android.os.RemoteException;
import android.util.*;
import java.util.regex.*;
import com.maxwell.net.ethernet.EthernetMonitor;
import android.os.UserHandle;
import android.os.SystemService;
import android.os.SystemClock;
import android.app.AlarmManager;
import com.maxwell.provisioning.util.DateAndTimeHelper;
/**
 * EthernetService handles remote Ethernet operation requests by implementing
 * the IEthernetManager interface. It also creates a EtherentMonitor to listen
 * for Etherent-related events.
 *
 * @hide
 */
public class EthernetService<syncronized> extends IEthernetManager.Stub {
    private static final String TAG = "EthernetService";
    private static final String ETHSTATE_CHANGE = "com.maxwell.net.ethernet.ETHSTATE_CHANGE";
    private static final int ETHERNET_HAS_CONFIG = 1;
    private static final boolean localLOGV = false;
    private static final int CONNECTED = 1;
    private static final int DISCONNECTED = 2;
    private static final int PHYUP = 3;
    private static final int FAILED = -1;
    private static final int MAX_DHCP_RETRY = 2;
    private static final String INTERFACE_NAME = "eth0";
    private boolean mStackConnected;
    private boolean mHWConnected;
    private boolean mInterfaceStopped;
    private NetworkHandler mNetworkTarget;
    private String mInterfaceName ;
    private DhcpInfoInternal mDhcpInfo = null;
    private DhcpInfoInternal mStaticInfo;
    private EthernetMonitor mMonitor = null;
    private INetworkManagementService mNwService;
    private int mEthState= EthernetManager.ETHERNET_STATE_UNKNOWN;
    private int mIntEthState = EthernetManager.ETHERNET_STATE_UNKNOWN; /*For handling UI related update*/
    private Context mContext;
    private String[] mDevName;
    private NetworkInfo mNetworkInfo = new NetworkInfo(ConnectivityManager.TYPE_ETHERNET, 0, "ETH", "");
    private BroadcastReceiver mEthernetMonitorReceiver;
    private LinkProperties mLinkProperties;
    private boolean mIsTearDownInvoked = false;
    private static int retryfail = 0;
    private boolean mEthernetSoftStopped = false;
    private static final String ETHERNET_IPV4 = "ipv4";
    private static final String ETHERNET_IPV6 = "ipv6";
    private static final int TIME_DELAY = 2000;
    private boolean isBooted = false;

    private AlarmManager mAlarmManager;
    private PendingIntent mDhcpRenewalIntent;
    private static final int DHCP_RENEW = 0;
    private static final String ACTION_DHCP_RENEW = "android.net.ethernet.DHCP_RENEW";

    //Used for sanity check on setting up renewal
    private static final int MIN_RENEWAL_TIME_SECS = 3600 ;  // 1 hour
    
    //Remember DHCP configuration from first request
    private DhcpInfoInternal mInfo;
 
    public EthernetService(Context context) {
        mContext = context;

        /* Create Thread Handler */
        HandlerThread threadHandler = new HandlerThread("Handler Thread");
        threadHandler.start();

        /* create Networhandler for getting 
         * NETLINK event like PHYUP/CONNECTED/DISCONNECTED from kernel and 
         * DHCP related events */
        mNetworkTarget = new NetworkHandler(threadHandler.getLooper());

        mMonitor = new EthernetMonitor(mContext);

        mAlarmManager = (AlarmManager)mContext.getSystemService(Context.ALARM_SERVICE);
        Intent dhcpRenewalIntent = new Intent(ACTION_DHCP_RENEW, null);
        mDhcpRenewalIntent = PendingIntent.getBroadcast(mContext, DHCP_RENEW, dhcpRenewalIntent, 0);

        mDhcpInfo = new DhcpInfoInternal();

        IBinder b = ServiceManager.getService(Context.NETWORKMANAGEMENT_SERVICE);
        mNwService = INetworkManagementService.Stub.asInterface(b);

        /*Register receiver for getting the polling result from kernel */
        IntentFilter filter = new IntentFilter();
        filter.addAction(EthernetManager.ETHERNET_MONITOR_RESULT);
        filter.addAction(Intent.ACTION_BOOT_COMPLETED);
        filter.addAction(ACTION_DHCP_RENEW);
	mEthernetMonitorReceiver = new EthernetMonitorReceiver();
	mContext.registerReceiver(mEthernetMonitorReceiver, filter);

    }

    /**
     * check if the ethernet service has been configured.
     * @return {@code true} if configured {@code false} otherwise
     */
    public boolean isConfigured() {
        final ContentResolver cr = mContext.getContentResolver();
        return (Settings.Secure.getInt(cr, Settings.Secure.ETHERNET_CONF, 0) == ETHERNET_HAS_CONFIG);

    }

    /**
     * Return the saved ethernet configuration
     * @return ethernet interface configuration on success, {@code null} on failure
     */
    public synchronized EthernetDevInfo getSavedConfig() {
        if (!isConfigured())
            return null;

        final ContentResolver cr = mContext.getContentResolver();
        EthernetDevInfo info = new EthernetDevInfo();
        String mode = Settings.Secure.getString(cr, Settings.Secure.ETHERNET_MODE);

        if(mode.equals(EthernetDevInfo.ETHERNET_CONN_MODE_DHCP)) {
            info.setConnectMode(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_MODE));
            info.setIfName(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_IFNAME));
            info.setNetType(ETHERNET_IPV4);
            info.setIpAddress(mDhcpInfo.ipAddress);
            info.setRouteAddr(SystemProperties.get("dhcp.eth0.gateway"));
            info.setDnsAddr(mDhcpInfo.dns1);
            info.setDns2Addr(mDhcpInfo.dns2);
            info.setNetMask(SystemProperties.get("dhcp.eth0.mask"));

            /* Vendor specific information (from RFC 2132)*/
            if(null != mDhcpInfo.vendorInfo && mDhcpInfo.vendorInfo.length() > 0) {
                info.setVendorInfo(mDhcpInfo.vendorInfo);
            }

            if(null != mDhcpInfo.provisioning_url && mDhcpInfo.provisioning_url.length() > 0) {
                info.setProvServerUrl(mDhcpInfo.provisioning_url);
            }

            if(null != mDhcpInfo.tftpserver_url && mDhcpInfo.tftpserver_url.length() > 0) {
                info.setTftpServerUrl(mDhcpInfo.tftpserver_url);
            }

            if(null != mDhcpInfo.sipserver_url && mDhcpInfo.sipserver_url.length() > 0) {
                info.setSipServerUrl(mDhcpInfo.sipserver_url);
            }

            if(null != mDhcpInfo.ntpserver_url && mDhcpInfo.ntpserver_url.length() > 0) {
                info.setNtpServerUrl(mDhcpInfo.ntpserver_url);
            }

        }else {
            info.setConnectMode(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_MODE));
            info.setNetType(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_NETTYPE));
            info.setIfName(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_IFNAME));
            info.setIpAddress(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_IP));
            info.setDnsAddr(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_DNS));
            info.setDns2Addr(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_DNS2));

            String nettype = getNetType();
            if(null != nettype && nettype.equals(ETHERNET_IPV4)) {
                info.setNetMask(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_MASK));
            } else if(null != nettype && nettype.equals(ETHERNET_IPV6)) {
                try {
                    info.setPrefixLength(Settings.Secure.getInt(cr, Settings.Secure.ETHERNET_PREFIXLENGTH));
                } catch (Exception e) {
                    Slog.e(TAG, "SettingNotFoundException for prefix length" + e);
                }
            } else {
                Slog.e(TAG, "Network Type(ipv4/ipv6) is not defined");
                return null;
            }
            info.setRouteAddr(Settings.Secure.getString(cr, Settings.Secure.ETHERNET_ROUTE));
        }

        return info;
    }

    /**
     * Set the ethernet interface configuration mode
     * @param mode {@code ETHERNET_CONN_MODE_DHCP} for dhcp {@code ETHERNET_CONN_MODE_MANUAL} for manual configure
     */
    public synchronized void setMode(String mode) {
        final ContentResolver cr = mContext.getContentResolver();
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_IFNAME, INTERFACE_NAME);
        Settings.Secure.putInt(cr, Settings.Secure.ETHERNET_CONF, 1);
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_MODE, mode);
    }

    /**
     * getMode
     * @return ethernet mode @dhcp or @manual
     */
    public synchronized String getMode() {
        final ContentResolver cr = mContext.getContentResolver();
        String mode = null;
        if(mEthState == EthernetManager.ETHERNET_STATE_ENABLED) {
            mode = Settings.Secure.getString(cr, Settings.Secure.ETHERNET_MODE);
        }
        return mode;
    }

    /**
     * getNetType
     * @return ethernet NetType @ipv4 or @ipv6
     */
    public synchronized String getNetType() {
        final ContentResolver cr = mContext.getContentResolver();
        String nettype = null;
        if(mEthState == EthernetManager.ETHERNET_STATE_ENABLED) {
            nettype = Settings.Secure.getString(cr, Settings.Secure.ETHERNET_NETTYPE);
        }
        return nettype;
    }

    /**
     * Set the ethernet nettype 
     * @param nettype {@ipv4} or {@ipv6}
     */
    public synchronized void setNetType(String nettype) {
        final ContentResolver cr = mContext.getContentResolver();
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_NETTYPE, nettype);
    }

    /**
     * update a ethernet interface information
     * @param info  the interface infomation
     */
    public synchronized boolean updateDevInfo(EthernetDevInfo info) {
        final ContentResolver cr = mContext.getContentResolver();
        Settings.Secure.putInt(cr, Settings.Secure.ETHERNET_CONF, 1);
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_NETTYPE, info.getNetType());
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_IFNAME, info.getIfName());
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_IP, info.getIpAddress());
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_MODE, info.getConnectMode());
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_ROUTE, info.getRouteAddr());
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_DNS, info.getDnsAddr());
        Settings.Secure.putString(cr, Settings.Secure.ETHERNET_DNS2, info.getDns2Addr());
        
        String nettype = getNetType();
        if(nettype.equals(ETHERNET_IPV4)) {
            Settings.Secure.putString(cr, Settings.Secure.ETHERNET_MASK, info.getNetMask());
        } else if(nettype.equals(ETHERNET_IPV6)) {
            Settings.Secure.putInt(cr, Settings.Secure.ETHERNET_PREFIXLENGTH, info.getPrefixLength());
        }

        if (mEthState == EthernetManager.ETHERNET_STATE_ENABLED) {
            try {
                return restartInterface();
            } catch (UnknownHostException e) {
                Slog.e(TAG, "Wrong ethernet configuration");
            }
        }
        return false;  
    }

    /**
     * resetInterface
     */
    public synchronized void resetInterface() {
        if (mEthState == EthernetManager.ETHERNET_STATE_ENABLED) {
            try {
                restartInterface();
            } catch(UnknownHostException e) {
                Slog.e(TAG, "Can not resetInterface");
            }
        }
    }

    /**
     * clearIPAddress
     */
    public synchronized void clearIPAddress() {
        if (mEthState == EthernetManager.ETHERNET_STATE_ENABLED) {
            try {
                clearAddress();
            } catch(UnknownHostException e) {
                Slog.e(TAG, "clear ipaddress is failed");
            }
        }
    }

    /**
     * get the number of ethernet interfaces in the system
     * @return the number of ethernet interfaces
     */
    public int getTotalInterface() {
        return EthernetNative.getInterfaceCnt();
    }


    private int scanDevice() {
        int i, j;
        if ((i = EthernetNative.getInterfaceCnt()) == 0)
            return 0;

        mDevName = new String[i];

        for (j = 0; j < i; j++) {
            mDevName[j] = EthernetNative.getInterfaceName(j);
            if (mDevName[j] == null)
                break;
            if (localLOGV) Slog.v(TAG, "device " + j + " name " + mDevName[j]);
        }

        return i;
    }

    /**
     * get all the ethernet device names
     * @return interface name list on success, {@code null} on failure
     */
    public String[] getDeviceNameList() {
        return (scanDevice() > 0) ? mDevName : null;
    }

    private int getPersistedState() {
        final ContentResolver cr = mContext.getContentResolver();
        try {
            return Settings.Secure.getInt(cr, Settings.Secure.ETHERNET_ON);
        } catch (Settings.SettingNotFoundException e) {
            return EthernetManager.ETHERNET_STATE_UNKNOWN;
        }
    }

    private int getPersistentIntermediateState() {
        final ContentResolver cr = mContext.getContentResolver();
        try {
            return Settings.Secure.getInt(cr, Settings.Secure.ETHERNET_INTERMEDIATE_STATE);
        } catch (Settings.SettingNotFoundException e) {
            return EthernetManager.ETHERNET_STATE_UNKNOWN;
        }
    }

    private synchronized void persistEnabled(boolean enabled) {
        final ContentResolver cr = mContext.getContentResolver();
        Settings.Secure.putInt(cr, Settings.Secure.ETHERNET_ON, enabled ? EthernetManager.ETHERNET_STATE_ENABLED : EthernetManager.ETHERNET_STATE_DISABLED);
    }

    private synchronized void setPersistentIntermediateState(int state) {
        final ContentResolver cr = mContext.getContentResolver();
        Settings.Secure.putInt(cr, Settings.Secure.ETHERNET_INTERMEDIATE_STATE, state);
    }


    /**
     * Enable or Disable a ethernet service
     * @param enable {@code true} to enable, {@code false} to disable
     */
    public synchronized void setState(int state) {

        Slog.d(TAG, "setState" + state + mEthState);
	
        if (state == mEthState) {
            return ;
        }

	if(state == EthernetManager.ETHERNET_STATE_ENABLED &&
			mEthState != EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED){
            Slog.i(TAG, "Ethernet enabled");

            if (EthernetNative.initEthernetNative() != 0) {
                Slog.e(TAG,"Can not init ethernet device layers");
            }

            if(mMonitor != null){
                mMonitor.startMonitoring();
            }else{
                Slog.e(TAG, "startMonitoring is failed");
            }

            if(!mEthernetSoftStopped) {

                NetworkUtils.resetConnections(INTERFACE_NAME, NetworkUtils.RESET_ALL_ADDRESSES);
                NetworkUtils.disableInterface(INTERFACE_NAME);

                try {
                    setInterfaceUp();
		} catch (UnknownHostException e) {
		    Slog.e(TAG, "Wrong to set interface up");
		}

		Slog.i(TAG, "Starting DSDT");
		SystemService.restart("dsdt");
		mEthernetSoftStopped = true;
            }

            mEthState = EthernetManager.ETHERNET_STATE_ENABLED;
	    persistEnabled(true);
	    mIntEthState = EthernetManager.ETHERNET_STATE_ENABLING;
	    sendEthStateBroadcast(mEthState, mIntEthState);
	    setPersistentIntermediateState(mIntEthState);

	    /*Post EthernetManager.EVENT_DEFERRED_PHY_STATE event after 2 second. 
	    This is required for knowing the cable status whether it is plugged-in or plugged-out
	    This is complaint with kernel polling event*/
	    mNetworkTarget.sendEmptyMessageDelayed(EthernetManager.EVENT_DEFERRED_PHY_STATE, 
				TIME_DELAY);

	}else if(state == EthernetManager.ETHERNET_STATE_DISABLED &&
			mEthState != EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED){
            Slog.i(TAG, "Ethernet disabled");
	    mEthState = EthernetManager.ETHERNET_STATE_DISABLED;
	    mIntEthState = EthernetManager.ETHERNET_STATE_DISABLING;
	    setPersistentIntermediateState(mIntEthState);
	    mIsTearDownInvoked = false;
	    persistEnabled(false);
	    sendEthStateBroadcast(mEthState, EthernetManager.ETHERNET_STATE_DISABLED);
	    /*Stopping the DSDT*/
	    Slog.i(TAG, "Stopping DSDT");
	    SystemService.stop("dsdt");
	    mEthernetSoftStopped = false;
	    stopInterface(false);
	    mMonitor.pauseMonitoring();
	    EthernetNative.deinitEthernetNative();

	}else if (state == EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED){
            Slog.i(TAG, "Ethernet PC port enabled ");

            if (null == getSavedConfig() || null == INTERFACE_NAME){
                Slog.i(TAG, "Ethernet PC port: Fresh");
		mNetworkInfo.setDetailedState(DetailedState.DISCONNECTED, null, null);
		mNetworkInfo.setIsAvailable(false);
		mEthState = EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED;
		sendEthStateChangeBroadcast(EthernetManager.EVENT_HW_DISCONNECTED);
            }else if(mEthState == EthernetManager.ETHERNET_STATE_ENABLED){
                mEthState = EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED;
                notifyStateChange(INTERFACE_NAME, DetailedState.DISCONNECTED);
	    }
            persistEnabled(true);
            mMonitor.pauseMonitoring();
            EthernetNative.deinitEthernetNative();
            mEthernetSoftStopped = true;

	}else if (state == EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_DISABLED){
            Slog.i(TAG, "Ethernet PC port disabled ");
            mEthState = EthernetManager.ETHERNET_STATE_DISABLED;
            setState(EthernetManager.ETHERNET_STATE_ENABLED);
	}
    }

    /**
     * Get ethernet service state
     * @return the state of the ethernet service
     */
    public int getState( ) {
        return mEthState;
    }

    /**
     * Get ethernet Intermediate state
     * @return the state of the ethernet service
     */
    public int getIntState( ) {
        return mIntEthState;
    }

    /**
     * Stop etherent interface
     * @param suspend {@code false} disable the interface {@code true} only reset the connection without disable the interface
     * @return true
     */
    private boolean stopInterface(boolean suspend) {
                synchronized (mNetworkTarget) {
                    mInterfaceStopped = true;
                    Slog.i(TAG, "stop dhcp and interface");
                    mNetworkTarget.removeMessages(EthernetManager.EVENT_DHCP_START);

                    mAlarmManager.cancel(mDhcpRenewalIntent);
                    if (!NetworkUtils.stopDhcp(INTERFACE_NAME)) {
                        Slog.w(TAG, "Could not stop DHCP");
                    }
                    NetworkUtils.resetConnections(INTERFACE_NAME, NetworkUtils.RESET_ALL_ADDRESSES);
                    if (!suspend) {
                        NetworkUtils.disableInterface(INTERFACE_NAME);
                    }

                    IBinder b = ServiceManager.getService(Context.NETWORKMANAGEMENT_SERVICE);
                    INetworkManagementService service = INetworkManagementService.Stub.asInterface(b);
                    try {
                        service.clearInterfaceAddresses(INTERFACE_NAME);
                    } catch (Exception e) {
                        Log.e(TAG, "Failed to clear addresses or disable ip" + e);
                    }

                    if(suspend == false) {
                        mIntEthState = EthernetManager.ETHERNET_STATE_DISABLED;
                        setPersistentIntermediateState(mIntEthState);
                        sendEthStateBroadcast(mEthState, mIntEthState);
                    } else if (getState() == EthernetManager.ETHERNET_STATE_ENABLED && suspend == true) {
                        mIntEthState = EthernetManager.ETHERNET_LINK_DOWN;
                        setPersistentIntermediateState(mIntEthState);
                        sendEthStateBroadcast(mEthState, mIntEthState);
                    }
                }
        return true;
    }

    private void setInterfaceUp() throws UnknownHostException {
                try {
                    mNwService.setInterfaceUp(INTERFACE_NAME);
		} catch (RemoteException re){
                    Slog.e(TAG,"Set interface up failed: " + re);
		} catch (IllegalStateException e) {
                    Slog.e(TAG,"Set interface up fialed: " + e);
		}
    }

    private boolean configureInterface(EthernetDevInfo info) throws UnknownHostException {
        DhcpInfo infoStatic = new DhcpInfo();
        mStackConnected = false;
        mInterfaceStopped = false;

        /*Reset All Addresses @ipv4 and @ipv6*/  
        //NetworkUtils.resetConnections(info.getIfName(), NetworkUtils.RESET_ALL_ADDRESSES);
        
        mInterfaceName = info.getIfName();

        if (info.getConnectMode().equals(EthernetDevInfo.ETHERNET_CONN_MODE_DHCP)) {  /*dhcp case*/
            if (localLOGV) Slog.i(TAG, "trigger dhcp for device " + info.getIfName());
		    mNetworkTarget.sendEmptyMessage(EthernetManager.EVENT_DHCP_START);
        } else { /*static case*/
            int event = -1;

            if(info.getNetType().equals(ETHERNET_IPV6)) {   /*IPV6 case*/
                mIntEthState = EthernetManager.ETHERNET_STATIC_IP_CONFIGURING;
                setPersistentIntermediateState(mIntEthState);

                //MAX-3941 : Added to avoid picking up google default dns after reboot
                //It will be checked in updateDns() function in ConnectivityService.java
                //dnses.size() == 0 && mDefaultDns != null
                mStaticInfo = new DhcpInfoInternal();
                mStaticInfo.ipAddress = info.getIpAddress();
                mStaticInfo.prefixLength = info.getPrefixLength();

                if(null != info.getRouteAddr()) {
                    InetAddress gatewayAddress = NetworkUtils.numericToInetAddress(info.getRouteAddr());
                    RouteInfo defaultRoute = new RouteInfo(gatewayAddress);
                    mStaticInfo.addRoute(defaultRoute);
                }
                mStaticInfo.dns1 = info.getDnsAddr();
                mStaticInfo.dns2 = info.getDns2Addr();

                mLinkProperties = mStaticInfo.makeLinkProperties();
                mLinkProperties.setInterfaceName(mInterfaceName);
            
                if (NetworkUtils.configureInterface6(info.getIfName(), info.getIpAddress(), info.getRouteAddr(), info.getDnsAddr(), 
                                                     info.getDns2Addr(), info.getPrefixLength())) {
                    event = EthernetManager.EVENT_INTERFACE_CONFIGURATION_SUCCEEDED;
                    Slog.d(TAG, "Static IP configuration succeeded");
                } else {
                    event = EthernetManager.EVENT_INTERFACE_CONFIGURATION_FAILED;
                    Slog.e(TAG, "Static IP configuration failed");
                    mNetworkTarget.sendEmptyMessage(event);
                    return false;
                }
            } else if(info.getNetType().equals(ETHERNET_IPV4)) {    /*IPV4 case*/

                mStaticInfo = new DhcpInfoInternal();
                mStaticInfo.ipAddress = info.getIpAddress();
                mStaticInfo.prefixLength = NetworkUtils.netmaskIntToPrefixLength(lookupHost(info.getNetMask()));

                LinkAddress linkAddress = new LinkAddress(Inet4Address.ANY, mStaticInfo.prefixLength);

                InetAddress gatewayAddress = NetworkUtils.intToInetAddress(lookupHost(info.getRouteAddr()));
                RouteInfo defaultRoute = new RouteInfo(gatewayAddress);
                mStaticInfo.addRoute(defaultRoute);
                mStaticInfo.dns1 = info.getDnsAddr();
                mStaticInfo.dns2 = info.getDns2Addr();

                mLinkProperties = mStaticInfo.makeLinkProperties();
                mLinkProperties.setInterfaceName(mInterfaceName);

                infoStatic = mStaticInfo.makeDhcpInfo();
                if (localLOGV) Slog.i(TAG, "set ip manually " + infoStatic.toString());

                mIntEthState = EthernetManager.ETHERNET_STATIC_IP_CONFIGURING;
                setPersistentIntermediateState(mIntEthState);

                if (NetworkUtils.configureInterface(info.getIfName(), infoStatic)) {
                    /*Reset ipv6 address if anything is there*/
                    NetworkUtils.resetConnections(info.getIfName(), NetworkUtils.RESET_IPV6_ADDRESSES);
                    event = EthernetManager.EVENT_INTERFACE_CONFIGURATION_SUCCEEDED;
                    Slog.d(TAG, "Static IP configuration succeeded");
                } else {
                    event = EthernetManager.EVENT_INTERFACE_CONFIGURATION_FAILED;
                    Slog.e(TAG, "Static IP configuration failed");
                    mNetworkTarget.sendEmptyMessage(event);
                    return false; 
                }
            }  
            mNetworkTarget.sendEmptyMessage(event);
        }
        return true;
    }

    /**
     * reset ethernet interface
     * @return true
     * @throws UnknownHostException
     */
    private boolean restartInterface()  throws UnknownHostException{
            EthernetDevInfo info = getSavedConfig();
            if (info != null && isConfigured()) {
                synchronized (this) {
                    mInterfaceName = info.getIfName();
                    if (localLOGV) Slog.i(TAG, "reset device " + mInterfaceName);
                    /* Stop DHCP*/
                    if (mNetworkTarget != null) {
                        mNetworkTarget.removeMessages(EthernetManager.EVENT_DHCP_START);
                    }
                    mAlarmManager.cancel(mDhcpRenewalIntent); 
                    if (!NetworkUtils.stopDhcp(mInterfaceName)) {
                        Slog.w(TAG, "Could not stop DHCP");
                    }
                    return configureInterface(info);
                }
            }
        return true;
    }

    /**
     * clear IP Address
     * @throws UnknownHostException
     */
    private boolean clearAddress()  throws UnknownHostException {
                synchronized (this) {
                    if (localLOGV) Slog.i(TAG, "reset device " + INTERFACE_NAME);

                    IBinder b = ServiceManager.getService(Context.NETWORKMANAGEMENT_SERVICE);
                    INetworkManagementService service = INetworkManagementService.Stub.asInterface(b);
                    try {
                        service.clearInterfaceAddresses(INTERFACE_NAME);
                    } catch (Exception e) {
                        Log.e(TAG, "Failed to clear addresses or disable ip" + e);
                    }
                }
        return true;
    }

    private void sendEthStateBroadcast(int ActState, int IntState){
        Intent intent = new Intent();
        Slog.v(TAG, "sendEthStateBroadcast:: isBooted " + isBooted);
        if(!isBooted) { //MAX-3975 : Added check to send broadcast before boot completed event is sent
            intent.addFlags(Intent.FLAG_RECEIVER_REGISTERED_ONLY_BEFORE_BOOT);
        }
        intent.setAction(ETHSTATE_CHANGE);
        intent.putExtra("ActualState", ActState);
        intent.putExtra("IntermediateState", IntState);
        mContext.sendBroadcastAsUser(intent, UserHandle.ALL);
    }

    /*Receiver for the System UI Notification */
    private void sendEthStateChangeBroadcast(int event){
        Intent intent = new Intent();
        if (localLOGV) Slog.v(TAG, "sendEthStateChangeBroadcast<<<");
        Slog.v(TAG, "sendEthStateChangeBroadcast:: isBooted " + isBooted);
        if(!isBooted) { //MAX-3975 : Added check to send broadcast before boot completed event is sent
            intent.addFlags(Intent.FLAG_RECEIVER_REGISTERED_ONLY_BEFORE_BOOT);
        }
        intent.setAction(EthernetManager.ETHERNET_STATE_CHANGED_ACTION);
        intent.putExtra("EthEvent", event);
        intent.putExtra(EthernetManager.EXTRA_NETWORK_INFO, new NetworkInfo(mNetworkInfo));
        intent.putExtra(EthernetManager.EXTRA_LINK_PROPERTIES, new LinkProperties(mLinkProperties));
        mContext.sendBroadcastAsUser(intent, UserHandle.ALL);
    }

    private void setNotify(DetailedState state, int event) {
        if (localLOGV) Slog.v(TAG, "setNotify :" + state + ",mNetworkInfo.getDetailedState():"+mNetworkInfo.getDetailedState());
        if (mNetworkInfo.getDetailedState() != state || getState() == EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED) {

            mNetworkInfo.setDetailedState(state, null, null);
            if(state == DetailedState.DISCONNECTED){
				stopInterface(true);
	            mNetworkInfo.setIsAvailable(false);
            }else{
				mNetworkInfo.setIsAvailable(true);
			}

            sendEthStateChangeBroadcast(event);
        }
    }

    public void notifyPhyConnected(String ifname) {
        if(mIsTearDownInvoked == false){
        if (localLOGV) Slog.v(TAG, "report interface is up for " + ifname);
        synchronized(this) {
            mIntEthState = EthernetManager.ETHERNET_STATE_ENABLED;
	    sendEthStateBroadcast(mEthState, mIntEthState);
	    setPersistentIntermediateState(mIntEthState);
            mNetworkTarget.sendEmptyMessage(EthernetManager.EVENT_HW_PHYCONNECTED);
        }
        }else{
            Slog.e(TAG, "notifyPhyConnected should not be called");
        }
    }

    public void notifyStateChange(String ifname, DetailedState state) {
        if(mIsTearDownInvoked == false){
            if (localLOGV) Slog.i(TAG, "report new state " + state.toString() + " on dev " + ifname);
			if(state.equals(DetailedState.DISCONNECTED)){
	            SystemProperties.set("dhcp." + ifname + ".ipaddress", "");
	            SystemProperties.set("dhcp." + ifname + ".gateway", "");
	            SystemProperties.set("dhcp." + ifname + ".dns1", "");
	            SystemProperties.set("dhcp." + ifname + ".dns2", "");
	            SystemProperties.set("dhcp." + ifname + ".mask", "");
	            mDhcpInfo.ipAddress = null;
	            mDhcpInfo.dns1 = null;
	            mDhcpInfo.dns2 = null;
	            mDhcpInfo.prefixLength = -1;
	        }
			/*If disconnect event comes while DHCP trial going, stop DHCP immidiately and update*/
			if(state.equals(DetailedState.DISCONNECTED) && mNetworkInfo.getDetailedState().equals(DetailedState.OBTAINING_IPADDR)){

				setNotify(DetailedState.DISCONNECTED, EthernetManager.EVENT_HW_DISCONNECTED);
				return;
			}

	        if (ifname.equals(INTERFACE_NAME)) {
            if (localLOGV) Slog.v(TAG, "update network state tracker");
            synchronized(this) {
                mNetworkTarget.sendEmptyMessage(state.equals(DetailedState.CONNECTED)
                        ? EthernetManager.EVENT_HW_CONNECTED : EthernetManager.EVENT_HW_DISCONNECTED);
            }
        }
       }else{
        Slog.i(TAG, "notifyStateChange should not be called");
       }
    }

    private static int lookupHost(String hostname) {
        InetAddress inetAddress;
        try {
            inetAddress = InetAddress.getByName(hostname);

        } catch (UnknownHostException e) {
            return -1;
        }
        byte[] addrBytes;
        int addr;
        addrBytes = inetAddress.getAddress();
        addr = ((addrBytes[3] & 0xff) << 24)
            | ((addrBytes[2] & 0xff) << 16)
            | ((addrBytes[1] & 0xff) << 8)
            |  (addrBytes[0] & 0xff);
        return addr;
    }

    public synchronized void checkAndStartEthernet(){
        Slog.i(TAG, "checkAndStartEthernet called:: isBooted " + isBooted);

        int isEnabled = getPersistedState();
        if (localLOGV) Slog.i(TAG, "Ethernet dev enabled " + isEnabled);

        if(isEnabled == EthernetManager.ETHERNET_STATE_UNKNOWN){
            Slog.i(TAG, "By default, ethernet is enabled");
            mEthState = mIntEthState = EthernetManager.ETHERNET_STATE_DISABLED;
            setPersistentIntermediateState(mIntEthState);
            sendEthStateBroadcast(mEthState, mIntEthState);
            
            /*Call this function to enable the ethernet at very first bootup*/
            setState(EthernetManager.ETHERNET_STATE_ENABLED);
        }else{
            setState(isEnabled);
        }
    }

    private void configureDhcpRenewal(DhcpInfoInternal dhcpInfo) {
        Slog.d(TAG, "configureDhcpRenewal");
        if(null != dhcpInfo) {
            long leaseDuration = dhcpInfo.leaseDuration; // int to long conversation
            Slog.d(TAG, "configureDhcpRenewal leaseDuration is " + leaseDuration);
            
            if(leaseDuration >= 0) {
                //TODO: would be good to notify the user that his network configuration is
                //bad and that the device cannot renew below MIN_RENEWAL_TIME_SECS   
                if (leaseDuration < MIN_RENEWAL_TIME_SECS) {
                    leaseDuration = MIN_RENEWAL_TIME_SECS;
                }
                //Do it a bit earlier than half the lease duration time
                //to beat the native DHCP client and avoid extra packets
                //48% for one hour lease time = 29 minutes
                mAlarmManager.set(AlarmManager.ELAPSED_REALTIME_WAKEUP,
                        SystemClock.elapsedRealtime() +
                        leaseDuration * 480, //in milliseconds
                        mDhcpRenewalIntent);  
            } else {
                //infinite lease time, no renewal needed
            }        
        }
    }

    private class NetworkHandler extends Handler {
        public NetworkHandler(Looper looper) {
            super(looper);
        }
         /*Note: mHWConnected shared between monitor thread and main looper thread, so should be synchrnized
		 *  mHWConnected is only modified in monitor receiver thread, not any other thread should modify it
		 **/
        public void handleMessage(Message msg) {
             
			if (localLOGV) Slog.i(TAG, "handleMessage: msg = " + msg.what);
			if (getState() == EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED )
			{
				if (EthernetManager.EVENT_HW_DISCONNECTED == msg.what){

					NetworkUtils.resetConnections(INTERFACE_NAME, NetworkUtils.RESET_ALL_ADDRESSES);
                    mAlarmManager.cancel(mDhcpRenewalIntent); 
					if (!NetworkUtils.stopDhcp(INTERFACE_NAME)) {
						if (localLOGV) Slog.w(TAG, "Could not stop DHCP");
					}
					synchronized (this){
					   mHWConnected = false;
					}
					setNotify(DetailedState.DISCONNECTED, msg.what);
				}else if (EthernetManager.EVENT_INTERFACE_CONFIGURATION_SUCCEEDED == msg.what){
                    mAlarmManager.cancel(mDhcpRenewalIntent);
					if (!NetworkUtils.stopDhcp(INTERFACE_NAME)) {
						if (localLOGV) Slog.w(TAG, "Could not stop DHCP");
					}
				}

			    this.removeMessages(EthernetManager.EVENT_DHCP_START);
				this.removeMessages(EthernetManager.EVENT_HW_DISCONNECTED);
				this.removeMessages(EthernetManager.EVENT_HW_CONNECTED);
				return;
			}

            switch (msg.what) {
               case EthernetManager.EVENT_DHCP_START:
                        int event = -1;
						Slog.d(TAG, "EthernetManager.EVENT_DHCP_START: cable not connected");
										
                        if (!mInterfaceStopped ) {
							synchronized (this){
								if (false == mHWConnected)
						           mInterfaceStopped = false;
							}
                            if (localLOGV) Slog.d(TAG, "DhcpHandler: DHCP request started");
							setNotify(DetailedState.OBTAINING_IPADDR, msg.what); 
                            mIntEthState = EthernetManager.ETHERNET_DHCP_IP_CONFIGURING;
                            setPersistentIntermediateState(mIntEthState);
                            sendEthStateBroadcast(mEthState, mIntEthState);
                            if (NetworkUtils.runDhcp(INTERFACE_NAME, mDhcpInfo)) {
                                event = EthernetManager.EVENT_INTERFACE_CONFIGURATION_SUCCEEDED;
                                Slog.d(TAG, "DhcpHandler: DHCP request succeeded: " + mDhcpInfo.toString());
                                configureDhcpRenewal(mDhcpInfo);
                                setDHCPNtpServer(mDhcpInfo.ntpserver_url);
				this.removeMessages(EthernetManager.EVENT_HW_PHYCONNECTED);
                                mInfo = mDhcpInfo;
                            } else {
                                event = EthernetManager.EVENT_DHCP_FAILED;
                                Slog.e(TAG, "DhcpHandler: DHCP request failed: " + NetworkUtils.getDhcpError());
                            }
                            mLinkProperties = mDhcpInfo.makeLinkProperties();
                            mLinkProperties.setInterfaceName(INTERFACE_NAME);
                            sendEmptyMessage(event);
                        } else {
                            mInterfaceStopped = false;
                        }
                        break;
                    case EthernetManager.EVENT_RENEW_DHCP:
                        Slog.d(TAG, "DhcpHandler: DHCP renewal request started: ");
						synchronized (this){
	                        if (mHWConnected && NetworkUtils.runDhcpRenew(INTERFACE_NAME, mDhcpInfo)) {
                            Slog.d(TAG, "DhcpHandler: DHCP renewal request succeeded: " + mDhcpInfo.toString());
                            mDhcpInfo.updateFromDhcpRequest(mInfo);
                            mLinkProperties = mDhcpInfo.makeLinkProperties();
                            configureDhcpRenewal(mDhcpInfo);
                            setDHCPNtpServer(mDhcpInfo.ntpserver_url);
                            this.removeMessages(EthernetManager.EVENT_HW_PHYCONNECTED);
                        } else {
                            Slog.e(TAG, "DhcpHandler: DHCP renewal request failed: " + NetworkUtils.getDhcpError());
                            sendEmptyMessage(EthernetManager.EVENT_DHCP_FAILED);
                        }
						}
                        break;  
                case EthernetManager.EVENT_INTERFACE_CONFIGURATION_SUCCEEDED:
                    if (localLOGV) Slog.i(TAG, "received configured succeeded, stack=" + mStackConnected + " HW=" + mHWConnected);
					synchronized (this){
	                    mStackConnected = true;
	                    retryfail = 0;
	                    if (mHWConnected) {
	                        setNotify(DetailedState.CONNECTED, msg.what);
							mIntEthState = EthernetManager.ETHERNET_IP_CONF_SUCCESS;
		                    setPersistentIntermediateState(mIntEthState);
		                    sendEthStateBroadcast(mEthState, mIntEthState);
	                    } 
					}                   
                    break;
                case EthernetManager.EVENT_INTERFACE_CONFIGURATION_FAILED:
					synchronized (this){
                    mStackConnected = false;
                    mIntEthState = EthernetManager.ETHERNET_STATIC_IP_ERROR;
                    setPersistentIntermediateState(mIntEthState);
                    sendEthStateBroadcast(mEthState, mIntEthState);
					}
                    break;
                case EthernetManager.EVENT_DHCP_FAILED:
					synchronized (this){
                    mStackConnected = false;
					if (!mHWConnected) /*Simply return*/
						return;
		            if (MAX_DHCP_RETRY >= ++retryfail){
		                Slog.e(TAG, "received configured failed.. Retrying");
                        sendEmptyMessage(EthernetManager.EVENT_DHCP_RECONNECT);
		            } else {
                                mIntEthState = EthernetManager.ETHERNET_DHCP_IP_ERROR;
                                setPersistentIntermediateState(mIntEthState);
                                sendEthStateBroadcast(mEthState, mIntEthState);
                                retryfail = 0;  
                            }
					}
                    break;
                case EthernetManager.EVENT_HW_CONNECTED:
                    if (localLOGV) Slog.i(TAG, "received HW connected, stack=" + mStackConnected + " HW=" + mHWConnected);
					synchronized (this){
                    if (mStackConnected) {
	                        setNotify(DetailedState.CONNECTED, msg.what);
                    }  
					}
                    break;
                case EthernetManager.EVENT_HW_DISCONNECTED:
                    if (localLOGV) Slog.i(TAG, "received disconnected events, stack=" + mStackConnected + " HW=" + mHWConnected);
					synchronized (this){
    					setNotify(DetailedState.DISCONNECTED, msg.what);
					}
                    break;
                case EthernetManager.EVENT_HW_PHYCONNECTED:
                    if(localLOGV) Slog.i(TAG, "interface up event, kick off connection request");
					synchronized (this){
	                    int state = getState();
	                    Slog.i(TAG, "interface up event, kick off connection request" + state);
	                    if (state != EthernetManager.ETHERNET_STATE_DISABLED) {
	                        setEthMode();
	                        EthernetDevInfo info = getSavedConfig();
	                        if (info != null && isConfigured()) {
	                            try {
	                                Slog.e(TAG, "configure interface");
	                                configureInterface(info);
	                            } catch (UnknownHostException e) {
	                                Slog.e(TAG, "Cannot configure interface");
	                            }
	                        }
	                    }
					}
                    break;
		case EthernetManager.EVENT_DHCP_RECONNECT:
		    Slog.i(TAG, "EVENT_DHCP_RECONNECT: trial = " + retryfail);
			synchronized (this){
			    if (MAX_DHCP_RETRY  >= retryfail){
	                        NetworkUtils.stopDhcp(INTERFACE_NAME);
				NetworkUtils.resetConnections(INTERFACE_NAME, NetworkUtils.RESET_ALL_ADDRESSES);
				try {
				    Slog.d(TAG, "retrying configure interface");
			            configureInterface(getSavedConfig());
				} catch (UnknownHostException e) {
				    Slog.e(TAG, "Cannot configure interface");
				}
			    } else{
				Slog.i(TAG, "Restarting ethenet at trial " + retryfail);
				setState(EthernetManager.ETHERNET_STATE_DISABLED);
				setState(EthernetManager.ETHERNET_STATE_ENABLED);
	                        retryfail = 0;
				}
			}
		    break;
                case EthernetManager.EVENT_DEFERRED_PHY_STATE:
					synchronized (this){
	                    int intermediateState = getIntState();
	                    Slog.i(TAG, "EVENT_DEFERRED_PHY_STATE" + intermediateState);

                    if(EthernetManager.ETHERNET_STATE_ENABLING == intermediateState) { 
                            Slog.i(TAG, "EEVENT_DEFERRED_PHY_STATE");
                            mIntEthState = EthernetManager.ETHERNET_LINK_DOWN;
                            setPersistentIntermediateState(EthernetManager.ETHERNET_LINK_DOWN);
                            sendEthStateBroadcast(mEthState, EthernetManager.ETHERNET_LINK_DOWN);
                    } 
					}
                    break;
	      }
       }
        
    }

    private void setDHCPNtpServer(String ntpServer) {
        Slog.i(TAG, "setDHCPNtpServer");
        if(null != ntpServer && ntpServer.length() > 0 && null != mContext) {
            Slog.i(TAG, "Setting DHCPNtpServer into android DB");
            DateAndTimeHelper helper = new DateAndTimeHelper(mContext);
            helper.setNtpServer("S_TIME_NTP_SERVER", ntpServer);
        } else {
            Slog.i(TAG, "Error caught during setDHCPNtpServer");
        }
    }

    //Receiver for getting the kernl event like PHYUP/CONNECTED/DISCONNECTED
    private class EthernetMonitorReceiver extends BroadcastReceiver {
        @Override
        public void onReceive(Context context, Intent intent) {
		    if (getState() == EthernetManager.ETHERNET_STATE_PCPORT_ETHERNET_ENABLED ) {
				return ;
		     }		 
            if(intent != null){
                if (intent.getAction().equals(EthernetManager.ETHERNET_MONITOR_RESULT)) {
                    int event = intent.getExtras().getInt("EventName");
                    String ifname = intent.getExtras().getString("InterfaceName");

                    Slog.d(TAG, "Event from netlink is" + event);
                    /*mNetworkTarget should be update holding mNetworkTarget object mutex*/ 
                    if(event == CONNECTED){
                        synchronized (mNetworkTarget){
						   mHWConnected = true;
						}
						notifyStateChange(ifname, DetailedState.CONNECTED);
                    }else if(event == DISCONNECTED){
                        synchronized (mNetworkTarget){
	                        mHWConnected = false;
	                    }
						notifyStateChange(ifname, DetailedState.DISCONNECTED);
                    }else if(event == PHYUP){
                        synchronized (mNetworkTarget){
						   mHWConnected = true;
						}
                        notifyPhyConnected(ifname);
                    }
                } else if(intent.getAction().equals(ACTION_DHCP_RENEW)) {
                    //DHCP renew
                    Log.d(TAG, "Sending a DHCP renewal " + this);
                    mNetworkTarget.sendEmptyMessage(EthernetManager.EVENT_RENEW_DHCP);
                } else if(intent.getAction().equals(Intent.ACTION_BOOT_COMPLETED)) {
                    Log.d(TAG, "Boot completed event is received:: isBooted " + isBooted);
                    isBooted = true;
                } 
            }
        }
    }

    private void setEthMode(){
        final ContentResolver cr = mContext.getContentResolver();
	String mode = Settings.Secure.getString(cr, Settings.Secure.ETHERNET_MODE);

	if(mode == null || mode.equals("") || mode.equals(EthernetDevInfo.ETHERNET_CONN_MODE_DHCP)) {
		setMode(EthernetDevInfo.ETHERNET_CONN_MODE_DHCP);
	}else {
		setMode(EthernetDevInfo.ETHERNET_CONN_MODE_MANUAL);
	}
    }

   /*
    * Request to reconnect already torn down ETH
    */
   public boolean startEth(){
       int state = getState();
       if(state == EthernetManager.ETHERNET_STATE_ENABLED && mIsTearDownInvoked == true){
           mIsTearDownInvoked = false;
           notifyPhyConnected(INTERFACE_NAME);
           return true;
       }
       return false;
   }

   /*
    * Request to tear down the ETH
    */
   public boolean stopEth(){
       int state = getState();
       if(state == EthernetManager.ETHERNET_STATE_ENABLED){
           notifyStateChange(INTERFACE_NAME, DetailedState.DISCONNECTED);
           mIsTearDownInvoked = true;
           return true;
       }
       return false;
   }

   /**
     * It is a silent way of restarting dsdt service alone
     * Note :: Right now, it is used in vlan scenarios
     * If vlan parameters are modified in,
     *   1) Ethernet is on in UI, but cable is unplugged
     */
   public boolean restartDSDT() {
       SystemService.stop("dsdt");
       if(null != mMonitor) {
           mMonitor.pauseMonitoring();
           mMonitor.startMonitoring();
       }
       SystemService.start("dsdt");
       return true;
   }

}
