/*

 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/gpio.h>
#include <linux/maxwell_poe.h>

struct maxwell_poe_data {
	struct platform_device	*pdev;
	struct maxwell_poe *poe;
	struct work_struct	poe_work;
};
char *inactive[2] = { "POE_STATE=INACTIVE", NULL };
char *active[2]    = { "POE_STATE=ACTIVE", NULL };

static void maxwell_poe_notify(struct work_struct *ws)
{
	char **uevent_envp = NULL;
	int value;

	struct maxwell_poe_data *poe_data =
		container_of(ws, struct maxwell_poe_data, poe_work);

	sysfs_notify(&poe_data->pdev->dev.kobj, NULL, "poe_status");
	printk("PoE Notify uEvent \n");
	value = gpio_get_value(poe_data->poe->gpio);
	uevent_envp = value ? inactive : active;
	kobject_uevent_env(&poe_data->pdev->dev.kobj, KOBJ_CHANGE, uevent_envp);
}

static irqreturn_t maxwell_poe_irq_handler(int irq, void *dev_id)
{
	struct maxwell_poe_data *poe_data = dev_id;

	schedule_work(&poe_data->poe_work);

	return IRQ_NONE;
}

static ssize_t show_poe_status(struct device *dev,
		struct device_attribute *attr, char *buf)
{

	struct maxwell_poe_data *poe_data = dev_get_drvdata(dev);

	int value = gpio_get_value(poe_data->poe->gpio);

	return sprintf(buf, "%d\n", value);
}

static DEVICE_ATTR(poe_status, S_IRUGO, show_poe_status, NULL);

static int poe_init(struct maxwell_poe_data *poe_data,
		struct maxwell_poe *poe)
{
	int err;
	int poe_irq;
	struct platform_device *pdev = poe_data->pdev;

	poe_data->poe = poe;

	err = gpio_request(poe->gpio, "Maxwell-PoE");
	if (err)
		return err;

	err = gpio_direction_input(poe->gpio);
	if (err)
		goto err_free_gpio;

	gpio_export(poe->gpio, 0);
	err = device_create_file(&pdev->dev, &dev_attr_poe_status);
	if (err)
		goto err_free_gpio;

	poe_irq = gpio_to_irq(poe->gpio);
	if (poe_irq < 0)
		return 0;

	INIT_WORK(&poe_data->poe_work, maxwell_poe_notify);
	irq_set_irq_type(poe_irq, IRQ_TYPE_EDGE_BOTH);
	err = request_irq(poe_irq, maxwell_poe_irq_handler, IRQF_SHARED,
			"maxwell-PoE-IRQ", poe_data);
	if (err)
		goto err_free_sysfs;

	return 0;

err_free_sysfs:
	device_remove_file(&pdev->dev, &dev_attr_poe_status);
err_free_gpio:
	gpio_free(poe->gpio);

	return err;
}

static void poe_free(struct maxwell_poe_data *poe_data)
{
	struct platform_device *pdev = poe_data->pdev;
	int poe_irq = gpio_to_irq(poe_data->poe->gpio);

	if (poe_irq >= 0)
		free_irq(poe_irq, poe_data);
	device_remove_file(&pdev->dev, &dev_attr_poe_status);
	gpio_free(poe_data->poe->gpio);
}

static ssize_t show_name(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "maxwell-poe\n");
}

static DEVICE_ATTR(name, S_IRUGO, show_name, NULL);

static int __devinit maxwell_poe_probe(struct platform_device *pdev)
{
	int err;
	struct maxwell_poe_data *poe_data;
	struct maxwell_poe_platform_data *pdata = pdev->dev.platform_data;

	if (!pdata)
		return -EINVAL;

	poe_data = kzalloc(sizeof(struct maxwell_poe_data), GFP_KERNEL);
	if (!poe_data)
		return -ENOMEM;

	poe_data->pdev = pdev;
	platform_set_drvdata(pdev, poe_data);

	if (pdata->poe) {
		err = poe_init(poe_data, pdata->poe);
		if (err)
			goto err_free_data;
	}

	err = device_create_file(&pdev->dev, &dev_attr_name);
	if (err)
		goto err_remove_name;


	dev_info(&pdev->dev, "GPIO fan initialized\n");

	return 0;

err_remove_name:
	device_remove_file(&pdev->dev, &dev_attr_name);
err_free_data:
	poe_free(poe_data);
	platform_set_drvdata(pdev, NULL);
	kfree(poe_data);

	return err;
}

static int __devexit maxwell_poe_remove(struct platform_device *pdev)
{
	struct maxwell_poe_data *poe_data = platform_get_drvdata(pdev);

	device_remove_file(&pdev->dev, &dev_attr_name);
	if (poe_data->poe)
		poe_free(poe_data);

	kfree(poe_data);

	return 0;
}

#ifdef CONFIG_PM
static int maxwell_poe_suspend(struct platform_device *pdev, pm_message_t state)
{
	//struct maxwell_poe_data *poe_data = platform_get_drvdata(pdev);


	return 0;
}

static int maxwell_poe_resume(struct platform_device *pdev)
{
	//struct maxwell_poe_data *poe_data = platform_get_drvdata(pdev);

	return 0;
}
#else
#define maxwell_poe_suspend NULL
#define maxwell_poe_resume NULL
#endif

static struct platform_driver maxwell_poe_driver = {
	.probe		= maxwell_poe_probe,
	.remove		= __devexit_p(maxwell_poe_remove),
	.suspend	= maxwell_poe_suspend,
	.resume		= maxwell_poe_resume,
	.driver	= {
		.name	= "maxwell-poe",
	},
};

static int __init maxwell_poe_init(void)
{
	return platform_driver_register(&maxwell_poe_driver);
}

static void __exit maxwell_poe_exit(void)
{
	platform_driver_unregister(&maxwell_poe_driver);
}

module_init(maxwell_poe_init);
module_exit(maxwell_poe_exit);

MODULE_DESCRIPTION("MAXWELL PoE driver");
MODULE_LICENSE("GPL");

